﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using CacheManager.Core.Configuration;

namespace CacheManager.Core.Cache
{
    public sealed class BaseCacheManager<TCacheValue> : BaseCache<TCacheValue>, ICacheManager<TCacheValue>
    {
        private ICacheHandle<TCacheValue>[] cacheHandles = new ICacheHandle<TCacheValue>[] { };

        private ICacheBackPlate cacheBackPlate = null;

        public event EventHandler<CacheActionEventArgs> OnRemove;

        public event EventHandler<CacheActionEventArgs> OnAdd;

        public event EventHandler<CacheActionEventArgs> OnPut;

        public event EventHandler<CacheUpdateEventArgs> OnUpdate;

        public event EventHandler<CacheActionEventArgs> OnGet;

        public event EventHandler<CacheClearEventArgs> OnClear;

        public event EventHandler<CacheClearRegionEventArgs> OnClearRegion;

        public ICacheManagerConfiguration Configuration
        {
            get;
            private set;
        }

        public IReadOnlyCollection<ICacheHandle<TCacheValue>> CacheHandles
        {
            get
            {
                return new ReadOnlyCollection<ICacheHandle<TCacheValue>>(new List<ICacheHandle<TCacheValue>>(cacheHandles));
            }
        }

        private BaseCacheManager(ICacheManagerConfiguration configuration)
        {
            if (configuration == null)
            {
                throw new ArgumentNullException("configuration");
            }

            this.Configuration = configuration;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="BaseCacheManager{TCacheValue}"/> class using the specified configuration and list of handles.
        /// <para>
        /// In this case, the <paramref name="configuration"/> will not get interpreted, only the handles passed in by 
        /// <paramref name="handles"/> will be added to the cache.
        /// The <paramref name="configuration"/> only defines the name of this instance.
        /// </para>
        /// </summary>
        /// <remarks>
        /// This constructor is primarily used for unit testing. To construct a cache manager, use the <see cref="CacheFactory"/>!
        /// </remarks>
        /// <param name="configuration">The configuration which defines the name of the manager.</param>
        /// <param name="handles">The list of cache handles.</param>
        public BaseCacheManager(ICacheManagerConfiguration configuration, params ICacheHandle<TCacheValue>[] handles)
            : this(configuration)
        {
            if (handles == null)
            {
                throw new ArgumentNullException("handles");
            }

            foreach (var handle in handles)
            {
                this.AddCacheHandle(handle);
            }
        }

        internal void SetCacheBackPlate(ICacheBackPlate backPlate)
        {
            if (backPlate == null)
            {
                throw new ArgumentNullException("backPlate");
            }
            
            this.cacheBackPlate = backPlate;

            var handles = new Func<ICacheHandle<TCacheValue>[]>(() =>
            {
                var handleList = new List<ICacheHandle<TCacheValue>>();
                foreach (var handle in this.cacheHandles)
                {
                    if (!handle.Configuration.IsBackPlateSource)
                    {
                        handleList.Add(handle);
                    }
                }
                return handleList.ToArray();
            });

            backPlate.SubscribeChanged((key) =>
            {
                this.EvictFromHandles(key, null, handles());
            });

            backPlate.SubscribeChanged((key, region) =>
            {
                this.EvictFromHandles(key, region, handles());
            });

            backPlate.SubscribeRemove((key) =>
            {
                this.EvictFromHandles(key, null, handles());
            });

            backPlate.SubscribeRemove((key, region) =>
            {
                this.EvictFromHandles(key, region, handles());
            });

            backPlate.SubscribeClear(() =>
            {
                this.ClearHandles(handles());
            });

            backPlate.SubscribeClearRegion((region) =>
            {
                this.ClearRegionHandles(region, handles());
            });
        }

        /// <summary>
        /// Adds a cache handle to the cache manager instance.
        /// </summary>
        /// <param name="handle">The cache handle.</param>
        public void AddCacheHandle(ICacheHandle<TCacheValue> handle)
        {
            if (handle == null)
            {
                throw new ArgumentNullException("handle");
            }

            var handleList = new List<ICacheHandle<TCacheValue>>(this.cacheHandles);
            handleList.Add(handle);
            this.cacheHandles = handleList.ToArray();
        }

        protected internal override bool AddInternal(CacheItem<TCacheValue> item)
        {
            if (item == null)
            {
                throw new ArgumentNullException("item");
            }

            var result = false;
            foreach (var handle in cacheHandles)
            {
                if (handle.Add(item))
                {
                    result = true;
                    handle.Stats.OnAdd(item);
                }
            }

            // trigger only once and not per handle and only if the item was added!
            if (result)
            {
                this.TriggerOnAdd(item.Key, item.Region);
            }

            return result;
        }

        protected internal override void PutInternal(CacheItem<TCacheValue> item)
        {
            if (item == null)
            {
                throw new ArgumentNullException("item");
            }

            foreach (var handle in cacheHandles)
            {
                if (handle.Configuration.EnableStatistics)
                {
                    // check if it is really a new item otherwise the items count is crap
                    // because we count it every time, but use only the current handle to 
                    // retrieve the item, otherwise we would trigger gets and find it in another handle maybe
                    var oldItem = string.IsNullOrWhiteSpace(item.Region) ?
                        handle.GetCacheItem(item.Key) :
                        handle.GetCacheItem(item.Key, item.Region);

                    handle.Stats.OnPut(item, oldItem == null);
                }

                handle.Put(item);
            }

            // update back plate
            if (this.cacheBackPlate != null)
            {
                if (string.IsNullOrWhiteSpace(item.Region))
                {
                    this.cacheBackPlate.NotifyChange(item.Key);
                }
                else
                {
                    this.cacheBackPlate.NotifyChange(item.Key, item.Region);
                }
            }

            this.TriggerOnPut(item.Key, item.Region);
        }

        protected override bool RemoveInternal(string key)
        {
            return this.RemoveInternal(key, null);
        }

        protected override bool RemoveInternal(string key, string region)
        {
            var result = false;

            foreach (var handle in cacheHandles)
            {
                var handleResult = false;
                if (!string.IsNullOrWhiteSpace(region))
                {
                    handleResult = handle.Remove(key, region);
                }
                else
                {
                    handleResult = handle.Remove(key);
                }

                if (handleResult)
                {
                    result = true;
                    handle.Stats.OnRemove(region);
                }
            }

            // update back plate
            if (this.cacheBackPlate != null)
            {
                if (string.IsNullOrWhiteSpace(region))
                {
                    this.cacheBackPlate.NotifyRemove(key);
                }
                else
                {
                    this.cacheBackPlate.NotifyRemove(key, region);
                }
            }

            // trigger only once and not per handle
            if (result)
            {
                this.TriggerOnRemove(key, region);
            }

            return result;
        }

        protected override CacheItem<TCacheValue> GetCacheItemInternal(string key)
        {
            return this.GetCacheItemInternal(key, null);
        }

        protected override CacheItem<TCacheValue> GetCacheItemInternal(string key, string region)
        {
            CacheItem<TCacheValue> cacheItem = null;

            for (int handleIndex = 0; handleIndex < this.cacheHandles.Length; handleIndex++)
            {
                var handle = cacheHandles[handleIndex];
                if (string.IsNullOrWhiteSpace(region))
                {
                    cacheItem = handle.GetCacheItem(key);
                }
                else
                {
                    cacheItem = handle.GetCacheItem(key, region);
                }

                handle.Stats.OnGet(region);

                if (cacheItem != null)
                {
                    // update last accessed, might be used for custom sliding implementations
                    cacheItem.LastAccessedUtc = DateTime.UtcNow;

                    // update other handles if needed
                    this.AddToHandles(cacheItem, handleIndex);
                    handle.Stats.OnHit(region);
                    this.TriggerOnGet(key, region);
                    break;
                }
                else
                {
                    handle.Stats.OnMiss(region);
                }
            }

            return cacheItem;
        }
        public bool Update(string key, Func<TCacheValue, TCacheValue> updateValue)
        {
            return Update(key, updateValue, new UpdateItemConfig());
        }

        public bool Update(string key, string region, Func<TCacheValue, TCacheValue> updateValue)
        {
            return Update(key, region, updateValue, new UpdateItemConfig());
        }

        public bool Update(string key, Func<TCacheValue, TCacheValue> updateValue, UpdateItemConfig config)
        {
            if (string.IsNullOrWhiteSpace(key))
            {
                throw new ArgumentNullException("key");
            }
            if (updateValue == null)
            {
                throw new ArgumentNullException("updateValue");
            }
            if (config == null)
            {
                throw new ArgumentNullException("config");
            }

            return UpdateInternal(key, updateValue, config);
        }

        public bool Update(string key, string region, Func<TCacheValue, TCacheValue> updateValue, UpdateItemConfig config)
        {
            if (string.IsNullOrWhiteSpace(key))
            {
                throw new ArgumentNullException("key");
            }

            if (string.IsNullOrWhiteSpace(region))
            {
                throw new ArgumentNullException("region");
            }
            if (updateValue == null)
            {
                throw new ArgumentNullException("updateValue");
            }
            if (config == null)
            {
                throw new ArgumentNullException("config");
            }

            return this.UpdateInternal(key, region, updateValue, config);
        }

        private bool UpdateInternal(string key, Func<TCacheValue, TCacheValue> updateValue, UpdateItemConfig config)
        {
            return UpdateInternal(key, null, updateValue, config);
        }

        private bool UpdateInternal(string key, string region, Func<TCacheValue, TCacheValue> updateValue, UpdateItemConfig config)
        {
            bool overallResult = false;
            bool overallVersionConflictOccurred = false;
            int overallTries = 1;

            for (int handleIndex = 0; handleIndex < this.cacheHandles.Length; handleIndex++)
            {
                var handle = this.cacheHandles[handleIndex];
                UpdateItemResult result;
                if (string.IsNullOrWhiteSpace(region))
                {
                    result = handle.Update(key, updateValue, config);
                }
                else
                {
                    result = handle.Update(key, region, updateValue, config);
                }

                if (result.Success)
                {
                    overallResult = true;
                    handle.Stats.OnUpdate(key, region, result);
                }

                if (result.VersionConflictOccurred)
                {
                    overallVersionConflictOccurred = true;
                }

                overallTries += result.NumberOfRetriesNeeded > 1 ? result.NumberOfRetriesNeeded - 1 : 0;

                if (result.VersionConflictOccurred && config.VersionConflictOperation != VersionConflictHandling.Ignore)
                {
                    if (!result.Success)
                    {
                        TriggerOnUpdate(key, region, config, new UpdateItemResult(overallVersionConflictOccurred, false, overallTries));
                        return false;
                    }

                    switch (config.VersionConflictOperation)
                    {
                        case VersionConflictHandling.EvictItemFromOtherCaches:
                            this.EvictFromOtherHandles(key, region, handleIndex);
                            break;
                        case VersionConflictHandling.UpdateOtherCaches:
                            CacheItem<TCacheValue> item;
                            if (string.IsNullOrWhiteSpace(region))
                            {
                                item = handle.GetCacheItem(key);
                            }
                            else
                            {
                                item = handle.GetCacheItem(key, region);
                            }

                            this.UpdateOtherHandles(item, handleIndex);
                            break;
                    }

                    // stop loop because we already handled everything.
                    break;
                }
            }

            // update back plate
            if (overallResult && this.cacheBackPlate != null)
            {
                if (string.IsNullOrWhiteSpace(region))
                {
                    this.cacheBackPlate.NotifyChange(key);
                }
                else
                {
                    this.cacheBackPlate.NotifyChange(key, region);
                }
            }

            TriggerOnUpdate(key, region, config, new UpdateItemResult(overallVersionConflictOccurred, overallResult, overallTries));

            return overallResult;
        }
        
        public override void Clear()
        {
            foreach (var handle in this.cacheHandles)
            {
                handle.Clear();
                handle.Stats.OnClear();
            } 
            
            if (this.cacheBackPlate != null)
            {
                this.cacheBackPlate.NotifyClear();
            }

            this.TriggerOnClear();
        }

        private void ClearOtherHandles(int excludeIndex)
        {
            if (excludeIndex < 0 || excludeIndex >= this.cacheHandles.Length)
            {
                throw new ArgumentOutOfRangeException("excludeIndex");
            }

            for (int handleIndex = 0; handleIndex < this.cacheHandles.Length; handleIndex++)
            {
                if (handleIndex != excludeIndex)
                {
                    var handle = this.cacheHandles[handleIndex];
                    handle.Clear();
                    handle.Stats.OnClear();
                }
            }

            this.TriggerOnClear();
        }

        private void ClearHandles(ICacheHandle<TCacheValue>[] handles)
        {
            foreach (var handle in handles)
            {
                handle.Clear();
                handle.Stats.OnClear();
            }

            this.TriggerOnClear();
        }

        public override void ClearRegion(string region)
        {
            if (string.IsNullOrWhiteSpace(region))
            {
                throw new ArgumentNullException("region");
            }

            foreach (var handle in this.cacheHandles)
            {
                handle.ClearRegion(region);
                handle.Stats.OnClearRegion(region);
            }

            if (this.cacheBackPlate != null)
            {
                this.cacheBackPlate.NotifyClearRegion(region);
            }

            this.TriggerOnClearRegion(region);
        }

        private void ClearRegionOtherHandles(string region, int excludeIndex)
        {
            if (excludeIndex < 0 || excludeIndex >= this.cacheHandles.Length)
            {
                throw new ArgumentOutOfRangeException("excludeIndex");
            }

            for (int handleIndex = 0; handleIndex < this.cacheHandles.Length; handleIndex++)
            {
                if (handleIndex != excludeIndex)
                {
                    var handle = this.cacheHandles[handleIndex];
                    handle.ClearRegion(region);
                    handle.Stats.OnClearRegion(region);
                }
            }

            this.TriggerOnClearRegion(region);
        }

        private void ClearRegionHandles(string region, ICacheHandle<TCacheValue>[] handles)
        {
            foreach (var handle in handles)
            {
                handle.ClearRegion(region);
                handle.Stats.OnClearRegion(region);
            }

            this.TriggerOnClearRegion(region);
        }

        private void TriggerOnRemove(string key, string region)
        {
            if (string.IsNullOrWhiteSpace(key))
            {
                throw new ArgumentNullException("key");
            }

            if (OnRemove != null)
            {
                this.OnRemove(this, new CacheActionEventArgs(key, region));
            }
        }

        private void TriggerOnAdd(string key, string region)
        {
            if (OnAdd != null)
            {
                this.OnAdd(this, new CacheActionEventArgs(key, region));
            }
        }

        private void TriggerOnPut(string key, string region)
        {
            if (OnPut != null)
            {
                this.OnPut(this, new CacheActionEventArgs(key, region));
            }
        }

        private void TriggerOnUpdate(string key, string region, UpdateItemConfig config, UpdateItemResult result)
        {
            if (OnUpdate != null)
            {
                this.OnUpdate(this, new CacheUpdateEventArgs(key, region, config, result));
            }
        }

        private void TriggerOnGet(string key, string region)
        {
            if (OnGet != null)
            {
                this.OnGet(this, new CacheActionEventArgs(key, region));
            }
        }

        private void TriggerOnClear()
        {
            if (OnClear != null)
            {
                this.OnClear(this, new CacheClearEventArgs());
            }
        }

        private void TriggerOnClearRegion(string region)
        {
            if (OnClearRegion != null)
            {
                this.OnClearRegion(this, new CacheClearRegionEventArgs(region));
            }
        }

        protected override void Dispose(bool disposeManaged)
        {
            if (disposeManaged)
            {
                foreach (var handle in this.cacheHandles)
                {
                    handle.Dispose();
                }
            }

            base.Dispose(disposeManaged);
        }

        private void UpdateOtherHandles(CacheItem<TCacheValue> item, int excludeIndex)
        {
            if (item == null)
            {
                return;
            }

            // .Where(p => p.Key != excludeIndex).Select(p => p.Value)
            for (int handleIndex = 0; handleIndex < this.cacheHandles.Length; handleIndex++)
            {
                if (handleIndex != excludeIndex)
                {
                    this.cacheHandles[handleIndex].Put(item);
                    //// handle.Stats.OnPut(item); don't update, 
                    //// we expect the item to be in the cache already at this point, so we should not increase the count...

                    this.TriggerOnPut(item.Key, item.Region);
                }
            }
        }

        private void EvictFromOtherHandles(string key, string region, int excludeIndex)
        {
            if (excludeIndex < 0 || excludeIndex >= this.cacheHandles.Length)
            {
                throw new ArgumentOutOfRangeException("excludeIndex");
            }

            for (int handleIndex = 0; handleIndex < this.cacheHandles.Length; handleIndex++)
            {
                if (handleIndex != excludeIndex)
                {
                    var handle = this.cacheHandles[handleIndex];
                    bool result;
                    if (string.IsNullOrWhiteSpace(region))
                    {
                        result = handle.Remove(key);
                    }
                    else
                    {
                        result = handle.Remove(key, region);
                    }

                    if (result)
                    {
                        handle.Stats.OnRemove(region);
                    }
                }
            }
        }

        private void EvictFromHandles(string key, string region, ICacheHandle<TCacheValue>[] handles)
        {
            foreach (var handle in handles)
            {
                bool result;
                if (string.IsNullOrWhiteSpace(region))
                {
                    result = handle.Remove(key);
                }
                else
                {
                    result = handle.Remove(key, region);
                }

                if (result)
                {
                    handle.Stats.OnRemove(region);
                }
            }
        }

        private void AddToHandles(CacheItem<TCacheValue> item, int foundIndex)
        {
            switch (this.Configuration.CacheUpdateMode)
            {
                case CacheUpdateMode.None:
                    // do basically nothing
                    break;

                case CacheUpdateMode.Full:
                    // update all cache handles except the one where we found the item
                    for (int handleIndex = 0; handleIndex < this.cacheHandles.Length; handleIndex++)
                    {
                        if (handleIndex != foundIndex)
                        {
                            this.cacheHandles[handleIndex].Add(item);
                        }
                    }
                    
                    break;

                case CacheUpdateMode.Up:
                    // optimizing so we don't even have to iterate
                    if (foundIndex == 0)
                    {
                        break;
                    }

                    // update all cache handles with lower order, up the list
                    for (int handleIndex = 0; handleIndex < this.cacheHandles.Length; handleIndex++)
                    {
                        if (handleIndex < foundIndex)
                        {
                            this.cacheHandles[handleIndex].Add(item);
                        }
                    }
                    
                    break;
            }
        }
    }
}